/* $Id: dbnj_hash_table.h,v 1.21 2005/10/24 20:17:23 oohara Exp $ */

#ifndef __LIBDBNJ_HASH_TABLE_H__
#define __LIBDBNJ_HASH_TABLE_H__

/* FILE */
#include <stdio.h>

#include "dbnj_list.h"

struct _dbnj_hash_node
{
  void *key;
  void (*key_delete_func)(void *);
  void *data;
  void (*data_delete_func)(void *);
};
typedef struct _dbnj_hash_node dbnj_hash_node;

struct _dbnj_hash_table
{
  /* number of key-data pairs in the table */
  int n_node;
  /* the size of the allocated memory */
  int size_node;
  dbnj_list **p_node;
  int (*key_hash_func)(const void *);
  /* return 1 (true) or 0 (false) */
  int (*key_equal_func)(const void *, const void *);
  void (*key_delete_func)(void *);
};
typedef struct _dbnj_hash_table dbnj_hash_table;

dbnj_hash_table *dbnj_hash_table_new(int n_errfile, FILE **p_errfile,
                                     int (*key_hash_func)(const void *),
                                     int (*key_equal_func)(const void *,
                                                           const void *),
                                     void (*key_delete_func)(void *));
void dbnj_hash_table_delete(int n_errfile, FILE **p_errfile,
                            dbnj_hash_table *table);
int dbnj_hash_table_insert(int n_errfile, FILE **p_errfile,
                           dbnj_hash_table *table, int replace,
                           void *key, void *data,
                           void (*data_delete_func)(void *));
int dbnj_hash_table_lookup(int n_errfile, FILE **p_errfile,
                           dbnj_hash_table *table,
                           const void *target_key,
                           void **found_key, void **found_data);
int dbnj_hash_table_remove(int n_errfile, FILE **p_errfile,
                           dbnj_hash_table *table,
                           const void *target_key);

#endif /* not __LIBDBNJ_HASH_TABLE_H__ */
