/* $Id: loop.c,v 1.69 2007/05/18 12:49:47 oohara Exp $ */

#include <stdio.h>
/* malloc */
#include <stdlib.h>
/* strlen, strerror */
#include <string.h>
/* errno */
#include <errno.h>

#include <SDL.h>
/* libsdl-ttf2.0 */
#include <SDL_ttf.h>

#include "dbnj_util.h"
#include "dbnj_sdl.h"
#include "markov.h"
#include "markov_lexical.h"
#include "world.h"

#include "loop.h"

static markov *read_doc_file(int n_errfile, FILE **p_errfile,
                             const char *doc_file_name);
static int draw_info(world *wp);

/* return
 * 0 if the game is over
 * 1 if the program should quit immediately
 */
int game_loop(world *wp, const char *doc_file_name)
{
  int i;
  int c;
  int n;
  int y;
  int column;
  int moved;
  int knock;
  int knock_max;
  int knock_word;
  int knock_word_max;
  int move_speed;
  int game_started;
  int flag_next;
  int hue;
  int successful;
  int rgb[3];
  SDL_Event event;
  markov *chain = NULL;
  const char *text_question = NULL;
  char *text_typed = NULL;
  char *text_typed_temp = NULL;
  const char *text_next = NULL;
  char buf[32];

  /* sanity check */
  if (wp == NULL)
  {
    fprintf(stderr, "game_loop: wp is NULL\n");
    return 1;
  }
  if (doc_file_name == NULL)
  {
    dbnj_fprintf(wp->n_errfile, wp->p_errfile,
                 "game_loop: doc_file_name is NULL");
    return 1;
  }

  chain = read_doc_file(wp->n_errfile, wp->p_errfile,
                        doc_file_name);
  if (chain == NULL)
  {
    dbnj_fprintf(wp->n_errfile, wp->p_errfile,
                 "game_loop: read_doc_file failed\n");
    return 1;
  }

  text_question = markov_next(wp->n_errfile, wp->p_errfile,
                                  chain);
  if (text_question == NULL)
  {
    dbnj_fprintf(wp->n_errfile, wp->p_errfile,
                 "game_loop: markov_next (#1) returned NULL\n");
    markov_delete(wp->n_errfile, wp->p_errfile,
                  chain);
    return 1;
  }
  text_next = markov_next(wp->n_errfile, wp->p_errfile,
                              chain);
  if (text_next == NULL)
  {
    dbnj_fprintf(wp->n_errfile, wp->p_errfile,
                 "game_loop: markov_next (#2) returned NULL\n");
    markov_delete(wp->n_errfile, wp->p_errfile,
                  chain);
    return 1;
  }

  column = 0;
  moved = 0;
  knock = 0;
  knock_max = 30;
  knock_word = 0;
  knock_word_max = 60;
  move_speed = 2;
  game_started = 0;
  text_typed = NULL;
  wp->char_done = 0;
  wp->char_done_cap = 250;
  wp->miss_10 = 0;

  if (wp->anija)
  {
    knock_max = 4;
    knock_word_max = 6;
    move_speed = 24;
    wp->char_done_cap = 1000;
  }

  wp->t_origin = SDL_GetTicks();
  while (1 == 1)
  {
    wp->t_now = SDL_GetTicks();
    if (!game_started)
      wp->t_origin = wp->t_now;
    flag_next = 0;

    while (SDL_PollEvent(&event))
    {
      switch(event.type)
      {
      case SDL_QUIT:
        markov_delete(wp->n_errfile, wp->p_errfile,
                      chain);
        return 1;
        break;
      case SDL_KEYDOWN:
        c = event.key.keysym.unicode;
        if (event.key.keysym.sym == SDLK_ESCAPE)
        {
          markov_delete(wp->n_errfile, wp->p_errfile,
                        chain);
          return 1;
        }
        else if ((c >= 32) && (c < 128))
        {
          game_started = 1;
          n = 0;
          if (text_typed != NULL)
            n = strlen(text_typed);
          if (c == ' ')
          {
            flag_next = 1;
          }
          else if ((text_question != NULL)
                   && (n < strlen(text_question))
                   && (c == text_question[n]))
          {
            buf[0] = c;
            buf[1] = '\0';
            text_typed_temp = dbnj_concat_string(wp->n_errfile, wp->p_errfile,
                                                 text_typed, 1, buf);
            if (text_typed_temp == NULL)
            {
              dbnj_fprintf(wp->n_errfile, wp->p_errfile,
                           "game_loop: can't add typed char '%c' (%d)\n",
                           c, c);
              if (text_typed != NULL)
              {
                free(text_typed);
                text_typed = NULL;
              }
              markov_delete(wp->n_errfile, wp->p_errfile,
                            chain);
              return 1;
            }
            text_typed = text_typed_temp;
            knock = knock_max;
          }
          else if ((text_question == NULL)
                   && (text_next != NULL)
                   && (c == text_next[0]))
          {
            if (text_typed != NULL)
            {
              free(text_typed);
              text_typed = NULL;
            }
            buf[0] = c;
            buf[1] = '\0';
            text_typed_temp = dbnj_concat_string(wp->n_errfile, wp->p_errfile,
                                                 text_typed, 1, buf);
            if (text_typed_temp == NULL)
            {
              dbnj_fprintf(wp->n_errfile, wp->p_errfile,
                           "game_loop: can't add typed char '%c' (%d)\n",
                           c, c);
              if (text_typed != NULL)
              {
                free(text_typed);
                text_typed = NULL;
              }
              markov_delete(wp->n_errfile, wp->p_errfile,
                            chain);
              return 1;
            }
            text_typed = text_typed_temp;
            knock = knock_max;
            text_question = text_next;
            text_next = markov_next(wp->n_errfile, wp->p_errfile,
                                    chain);
            if (text_next == NULL)
            {
              dbnj_fprintf(wp->n_errfile, wp->p_errfile,
                           "game_loop: markov_next returned NULL\n");
              if (text_typed != NULL)
              {
                free(text_typed);
                text_typed = NULL;
              }
              markov_delete(wp->n_errfile, wp->p_errfile,
                            chain);
              return 1;
            }
            column++;
            if (column >= 7)
              column = 0;
            moved = 0;
            knock_word = 0;
          }
        }
        break;
      default:
        break;
      }
    }

    if (wp->anija)
    {
      wp->char_done_cap = 1000;
    }
    else
    {
      wp->char_done_cap = ((wp->char_done / 250) + 1) * 250;
      if (wp->char_done_cap > 2000)
        wp->char_done_cap = 2000;
    }

    successful = 0;
    if ((text_question != NULL)
        && (text_typed != NULL)
        && (dbnj_string_equal(text_typed, text_question)))
      successful = 1;

    if ((text_question != NULL)
        && ((flag_next != 0)
            || (successful)
            || (moved > 480)))
    {
      if (successful)
      {
        (wp->miss_10)--;
        if (wp->miss_10 < 0)
          wp->miss_10 = 0;
        wp->char_done += strlen(text_question);
      }
      else
      {
        if (text_typed != NULL)
        {
          free(text_typed);
          text_typed = NULL;
        }
        wp->miss_10 += 10;
        if (wp->miss_10 >= 150)
          break;
      }
      text_question = NULL;
      moved = 0;
      knock_word = knock_word_max;
    }
    if ((flag_next != 0)
        || ((knock_word <= 0) && (text_question == NULL)))
    {
      if ((flag_next != 0) && (knock_word > 0) && (text_typed != NULL))
        wp->char_done += 2;

      text_question = text_next;
      if (text_typed != NULL)
      {
        free(text_typed);
        text_typed = NULL;
      }
      text_next = markov_next(wp->n_errfile, wp->p_errfile,
                              chain);
      if (text_next == NULL)
      {
        dbnj_fprintf(wp->n_errfile, wp->p_errfile,
                     "game_loop: markov_next returned NULL\n");
        if (text_typed != NULL)
        {
          free(text_typed);
          text_typed = NULL;
        }
        markov_delete(wp->n_errfile, wp->p_errfile,
                      chain);
        return 1;
      }
      column++;
      if (column >= 7)
        column = 0;
      moved = 0;
      knock = knock_max;
      knock_word = 0;
    }

    n = wp->char_done / 250;
    if (wp->anija)
      n = wp->char_done / 125;
    if (n < 0)
      n = 0;
    if (n > 7)
      n = 7;
    if (knock_max < 0)
      knock_max = 1;
    if (knock_word_max < 0)
      knock_word_max = 1;
    if (knock < -2)
      knock = -2;
    if (knock > knock_max)
      knock = knock_max;
    if (knock_word < 0)
      knock_word = 0;
    if (knock_word > knock_word_max)
      knock_word = knock_word_max;

    SDL_BlitSurface(wp->background[n], NULL, wp->screen, NULL);

    y = 10 + column * 30;
    if (text_question != NULL)
    {
      dbnj_draw_text(wp->n_errfile, wp->p_errfile,
                     wp->screen, wp->font_24, wp->antialias,
                     480 - moved, y, 0, text_question,
                     0, 0, 0, 47);
      dbnj_draw_text(wp->n_errfile, wp->p_errfile,
                     wp->screen, wp->font_24, wp->antialias,
                     0, y, 0, text_question,
                     255, 255, 255, 255);
    }
    if (text_typed != NULL)
    {
      hue = 180 + wp->miss_10;
      if ((knock <= 0) && (knock_word > 0))
      {
        dbnj_draw_text(wp->n_errfile, wp->p_errfile,
                       wp->screen, wp->font_24, wp->antialias,
                       0, y, 0, text_typed,
                       63, 63, 63, 255 * knock_word / knock_word_max);
      }
      else
      {
        for (i = 0; i < 3; i++)
          rgb[i] = 0;
        if (knock > 0)
        {  
          dbnj_hsv_to_rgb(wp->n_errfile, wp->p_errfile,
                          rgb, hue, 80 + 20 * knock / knock_max, 
                          25 + 75 * knock / knock_max);
        }
        else if (knock > -2)
        {
          rgb[0] = 255;
          rgb[1] = 255;
          rgb[2] = 255;
        }
        else
        {
          rgb[0] = 0;
          rgb[1] = 0;
          rgb[2] = 0;
        }
        dbnj_draw_text(wp->n_errfile, wp->p_errfile,
                       wp->screen, wp->font_24, wp->antialias,
                       0, y, 0, text_typed,
                       rgb[0], rgb[1], rgb[2], 255);
        if (strlen(text_typed) >= 2)
        {
          text_typed_temp = dbnj_concat_string(wp->n_errfile, wp->p_errfile,
                                               NULL, 1, text_typed);
          if (text_typed_temp == NULL)
          {
            dbnj_fprintf(wp->n_errfile, wp->p_errfile,
                         "game_loop: can't duplicate text_typed\n");
            if (text_typed != NULL)
            {
              free(text_typed);
              text_typed = NULL;
            }
            markov_delete(wp->n_errfile, wp->p_errfile,
                          chain);
            return 1;
          }
          text_typed_temp[strlen(text_typed) - 1] = '\0';
          dbnj_draw_text(wp->n_errfile, wp->p_errfile,
                         wp->screen, wp->font_24, wp->antialias,
                         0, y, 0, text_typed_temp,
                         0, 0, 0, 255);
          free(text_typed_temp);
          text_typed_temp = NULL;
        }
      }
    }
    if (text_next != NULL)
    {
      if (text_question != NULL)
      {
        rgb[0] = 231;
        rgb[1] = 231;
        rgb[2] = 231;
      }
      else
      {
        rgb[0] = 255;
        rgb[1] = 255;
        rgb[2] = 255;
      }
      dbnj_draw_text(wp->n_errfile, wp->p_errfile,
                     wp->screen, wp->font_24, wp->antialias,
                     0, (y + 30) % 210, 0, text_next,
                     rgb[0], rgb[1], rgb[2], 255);
    }

    /* maybe useful for debugging */
    /*
    if ((knock > 0) || (knock_word > 0))
    {
      char temp_buf[32];
      sprintf(temp_buf, "%d + %d", knock, knock_word);
      dbnj_draw_text(wp->n_errfile, wp->p_errfile,
                     wp->screen, wp->font_24, wp->antialias,
                     0, ((y + 30) % 210) + 30, 0, temp_buf,
                     0, 255, 0, 255);
    }
    */

    draw_info(wp);

    if (!game_started)
      dbnj_draw_text(wp->n_errfile, wp->p_errfile,
                     wp->screen, wp->font_12, wp->antialias,
                     240, 270, 2,
                     "press any key to start a game, or ESC to quit",
                     47, 49, 63, 255);

    SDL_Flip(wp->screen);

    if (game_started)
    {  
      (wp->frame)++;
      if ((wp->anija) && (wp->char_done >= 1000))
        break;
      if (wp->char_done >= 2000)
        break;
      if (wp->anija)
      {
        if (wp->char_done >= 750)
        {  
          knock_max = 4;
          knock_word_max = 4;
          move_speed = 480;
        }
        else if (wp->char_done >= 500)
        {  
          knock_max = 4;
          knock_word_max = 4;
          move_speed = 48;
        }
        else if (wp->char_done >= 250)
        {  
          knock_max = 4;
          knock_word_max = 5;
          move_speed = 36;
        }
        else if (wp->char_done >= 125)
        {  
          knock_max = 4;
          knock_word_max = 5;
          move_speed = 30;
        }
        else
        {  
          knock_max = 4;
          knock_word_max = 6;
          move_speed = 24;
        }
      }
      else if (wp->char_done >= 1750)
      {
        knock_max = 4;
        knock_word_max = 6;
        move_speed = 24;
      }
      else if (wp->char_done >= 1500)
      {
        knock_max = 6;
        knock_word_max = 9;
        move_speed = 24;
      }
      else if (wp->char_done >= 1250)
      {
        knock_max = 9;
        knock_word_max = 14;
        move_speed = 24;
      }
      else if (wp->char_done >= 1000)
      {
        knock_max = 14;
        knock_word_max = 20;
        move_speed = 24;
      }
      else if (wp->char_done >= 750)
      {
        knock_max = 20;
        knock_word_max = 30;
        move_speed = 24;
      }
      else if (wp->char_done >= 500)
      {
        knock_max = 30;
        knock_word_max = 60;
        move_speed = 24;
      }
      else if (wp->char_done >= 390)
      {
        knock_max = 30;
        knock_word_max = 60;
        move_speed = 12;
      }
      else if (wp->char_done >= 350)
      {
        knock_max = 30;
        knock_word_max = 60;
        move_speed = 8 + ((wp->char_done - 350) / 20) * 2;
      }
      else if (wp->char_done >= 250)
      {
        knock_max = 30;
        knock_word_max = 60;
        move_speed = 2 + (wp->char_done - 250) / 20;
      }
      else
      {
        knock_max = 30;
        knock_word_max = 60;
        move_speed = 2 + wp->char_done / 50;
      }

      if (knock > -2)
        knock--;
      else if (knock_word > 0)
        knock_word--;

      if ((text_question != NULL) && (knock <= 0) && (knock_word <= 0))
        moved += move_speed;
    }
    
    dbnj_sdl_wait(wp->t_now, 30, wp->clock_granularity);

  }

  if (text_typed != NULL)
  {
    free(text_typed);
    text_typed = NULL;
  }
  markov_delete(wp->n_errfile, wp->p_errfile,
                chain);
  return 0;
}

/* return NULL on error */
static markov *
read_doc_file(int n_errfile, FILE **p_errfile,
              const char *doc_file_name)
{
  char *s = NULL;
  markov *chain = NULL;
  FILE *doc_file_stream = NULL;
  void *yyscanner = NULL;

  /* sanity check */
  if (doc_file_name == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "read_doc_file: doc_file_name is NULL");
    return NULL;
  }

  chain = markov_new(n_errfile, p_errfile);
  if (chain == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "read_doc_file: markov_new failed\n");
    return NULL;
  }

  if (markov_lex_init(&yyscanner)!= 0)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "read_doc_file: markov_lex_init failed\n");
    markov_delete(n_errfile, p_errfile,
                  chain);
    return NULL;
  }

  errno = 0;
  doc_file_stream = fopen(doc_file_name, "r");
  if (doc_file_stream == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "read_doc_file: can't open file %s", doc_file_name);
    if ((errno != 0) && (strerror(errno) != NULL))
      dbnj_fprintf(n_errfile, p_errfile,
                   " (%s)", strerror(errno));
    dbnj_fprintf(n_errfile, p_errfile,
                 "\n");
    markov_lex_destroy(yyscanner);
    markov_delete(n_errfile, p_errfile,
                  chain);
    return NULL;
  }

  markov_read_from_file(doc_file_stream, yyscanner);
  s = NULL;
  if (markov_lex(n_errfile, p_errfile,
                 &chain, &s, yyscanner) != 0)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "read_doc_file: markov_lex failed\n");
    fclose(doc_file_stream);
    markov_lex_destroy(yyscanner);
    markov_delete(n_errfile, p_errfile,
                  chain);
    return NULL;
  }

  fclose(doc_file_stream);
  markov_lex_destroy(yyscanner);

  if (chain->n_state <= 0)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "read_doc_file: only one or no appropriate word found\n");
    markov_delete(n_errfile, p_errfile,
                  chain);
    return NULL;
  }

  return chain;
}

/* return 0 on success, 1 on error */
static int
draw_info(world *wp)
{
  int t;
  int red;
  int green;
  int blue;
  int status = 0;
  char buf[32];

  /* sanity check */
  if (wp == NULL)
  {
    dbnj_fprintf(wp->n_errfile, wp->p_errfile,
                 "draw_info: wp is NULL\n");
    return 1;
  }

  /*
  red = 63;
  green = 63;
  blue = 63;
  */
  red = 47;
  green = 49;
  blue = 63;

  if (wp->miss_10 < 150)
  {
    sprintf(buf, "miss %d.%1d", wp->miss_10 / 10, wp->miss_10 % 10);
    buf[31] = '\0';
    if (dbnj_draw_text(wp->n_errfile, wp->p_errfile,
                       wp->screen, wp->font_24, wp->antialias,
                       10, 300, 0, buf,
                       red, green, blue, 255) != 0)
    {
      dbnj_fprintf(wp->n_errfile, wp->p_errfile,
                   "draw_info: dbnj_draw_text (miss) failed\n");
      status = 1;
    }
  }

  sprintf(buf, "%d/%d", wp->char_done, wp->char_done_cap);
  buf[31] = '\0';
  if (dbnj_draw_text(wp->n_errfile, wp->p_errfile,
                     wp->screen, wp->font_24, wp->antialias,
                     10, 330, 0, buf,
                     red, green, blue, 255) != 0)
  {
    dbnj_fprintf(wp->n_errfile, wp->p_errfile,
                 "draw_info: dbnj_draw_text (char_done) failed\n");
    status = 1;
  }

  t = wp->t_now - wp->t_origin;
  if (t < 0)
  {
    dbnj_fprintf(wp->n_errfile, wp->p_errfile,
                 "draw_info: t is negative (%d)\n", t);
    status = 1;
    t = 0;
  }
  sprintf(buf, "%02d:%02d:%02d", t / (1000 * 60),
          (t / 1000) % 60, (t % 1000) / 10);
  buf[31] = '\0';
  if (dbnj_draw_text(wp->n_errfile, wp->p_errfile,
                     wp->screen, wp->font_24, wp->antialias,
                     470, 330, 1, buf,
                     red, green, blue, 255) != 0)
  {
    dbnj_fprintf(wp->n_errfile, wp->p_errfile,
                 "draw_info: dbnj_draw_text (time) failed\n");
    status = 1;
  }

  if ((t > 10000) && (wp->frame >= 500))
    sprintf(buf, "%d (%.1ffps)", wp->frame,
            ((double) wp->frame) * 1000.0 / ((double) t));
  else
    sprintf(buf, "%d (--.-fps)", wp->frame);
  buf[31] = '\0';
  if (dbnj_draw_text(wp->n_errfile, wp->p_errfile,
                     wp->screen, wp->font_12, wp->antialias,
                     470, 315, 1, buf,
                     red, green, blue, 255) != 0)
  {
    dbnj_fprintf(wp->n_errfile, wp->p_errfile,
                 "draw_info: dbnj_draw_text (fps) failed\n");
    status = 1;
  }

  if (wp->anija)
  {
    if (dbnj_draw_text(wp->n_errfile, wp->p_errfile,
                       wp->screen, wp->font_12, wp->antialias,
                       470, 300, 1, "--anija",
                       red, green, blue, 255) != 0)
    {
      dbnj_fprintf(wp->n_errfile, wp->p_errfile,
                   "draw_info: dbnj_draw_text (anija) failed\n");
      status = 1;
    }
  }

  return status;
}

void
game_over_loop(world *wp)
{
  int n;
  int t_temp;
  SDL_Event event;

  /* sanity check */
  if (wp == NULL)
  {
    fprintf(stderr, "game_over loop: wp is NULL\n");
    return;
  }

  wp->t_now = SDL_GetTicks();
  while (1 == 1)
  {
    t_temp = SDL_GetTicks();

    while (SDL_PollEvent(&event))
    {
      switch(event.type)
      {
      case SDL_QUIT:
        return;
        break;
      case SDL_KEYDOWN:
        if (event.key.keysym.sym == SDLK_ESCAPE)
          return;
        break;
      default:
        break;
      }
    }

    n = wp->char_done / 250;
    if (wp->anija)
      n = wp->char_done / 125;
    if (n < 0)
      n = 0;
    if (n > 7)
      n = 7;

    SDL_BlitSurface(wp->background[n], NULL, wp->screen, NULL);

    draw_info(wp);

    dbnj_draw_text(wp->n_errfile, wp->p_errfile,
                   wp->screen, wp->font_12, wp->antialias,
                   240, 270, 2,
                   "game over --- press ESC to quit",
                   47, 49, 63, 255);

    SDL_Flip(wp->screen);

    dbnj_sdl_wait(t_temp, 30, wp->clock_granularity);
  }
}
