/* $Id: main.c,v 1.49 2007/06/17 09:00:43 oohara Exp $ */

/* VERSION, DEFAULT_DOC, DEFAULT_FONT */
#include <config.h>

#include <stdio.h>
/* rand */
#include <stdlib.h>
/* time */
#include <time.h>

#include "dbnj_util.h"
#include "world.h"
#include "loop.h"

static void do_help(void);
static void do_version(void);

int
main(int argc, char *argv[])
{
  int i;
  const char *ttf_font_name = NULL;
  const char *doc_file_name = NULL;
  int antialias = 0;
  int anija = 0;
  int full_screen = 0;
  int flag_help = 0;
  int flag_version = 0;
  world *wp = NULL;

#ifdef DEFAULT_DOC
  doc_file_name = DEFAULT_DOC;
#endif /* DEFAULT_DOC */

#ifdef DEFAULT_FONT
  ttf_font_name = DEFAULT_FONT;
#endif /* DEFAULT_FONT */

  for (i = 1; i < argc; i++)
  {
    if (argv[i] == NULL)
      continue;

    if (dbnj_string_equal(argv[i], "--help"))
    {
      flag_help = 1;
    }
    else if (dbnj_string_equal(argv[i], "--version"))
    {
      flag_version = 1;
    }
    else if (dbnj_string_equal(argv[i], "--font"))
    {
      if (i + 1 < argc)
      {
        ttf_font_name = argv[i + 1];
      }
      else
      {
        fprintf(stderr, "main: missing argument for --font\n");
        return 1;
      }
      i += 1;
    }
    else if (dbnj_string_equal(argv[i], "--doc"))
    {
      if (i + 1 < argc)
      {
        doc_file_name = argv[i + 1];
      }
      else
      {
        fprintf(stderr, "main: missing argument for --doc\n");
        return 1;
      }
      i += 1;
    }
    else if (dbnj_string_equal(argv[i], "--full-screen"))
    {
      full_screen = 1;
    }
    else if (dbnj_string_equal(argv[i], "--anija"))
    {
      anija = 1;
    }
    else if (dbnj_string_equal(argv[i], "--antialias"))
    {
      antialias = 1;
    }
    else
    {
      fprintf(stderr, "main: unknown option: %s\n", argv[i]);
      fprintf(stderr, "Type %s --help for the list of options.\n", argv[0]);
      return 1;
    }
  }

  if ((flag_help != 0) || (flag_version != 0))
  {
    if (flag_help != 0)
      do_help();
    if (flag_version != 0)
      do_version();
    return 0;
  }

  if (ttf_font_name == NULL)
  {
    fprintf(stderr, "main: no TrueType font specified\n");
    return 1;
  }

  if (doc_file_name == NULL)
  {
    fprintf(stderr, "main: no document file specified\n");
    return 1;
  }

  srand((unsigned int) time(NULL));

  wp = world_new(ttf_font_name, antialias, anija, full_screen, 1);
  if (wp == NULL)
  {
    fprintf(stderr, "main: world_new failed\n");
    return 1;
  }

  if (game_loop(wp, doc_file_name) == 0)
    game_over_loop(wp);

  world_delete(wp);
  wp = NULL;

  return 0;
}

static void
do_help(void)
{
  printf("Usage: two-minute-hate [options]\n"
         "type-the-flying-word game with a knock-back feature\n"
         "Options:\n"
         );

  printf("  --doc _docname_    "
         "specify a text file for generating the word list\n");
#ifdef DEFAULT_DOC
  printf("    (default: %s)\n", DEFAULT_DOC);
#else /* not DEFAULT_DOC */
  printf("    (no default)\n");
#endif /* not DEFAULT_DOC */

  printf("  --font _fontname_  specify a font file\n");
#ifdef DEFAULT_FONT
  printf("    (default: %s)\n", DEFAULT_FONT);
#else /* not DEFAULT_FONT */
  printf("    (no default)\n");
#endif /* not DEFAULT_FONT */

  printf("  --full-screen      run the game in the full screen mode\n"
         );
  printf("  --anija            force maximum speed\n"
         "  --antialias        enable antialiasing a font\n");
  printf("  --help             print this message\n"
         "  --version          print version information\n"
         );
  printf("\n"
         "Report bugs to <oohara@libra.interq.or.jp>.\n"
         );
}

static void
do_version(void)
{
  printf("two-minute-hate %s\n", VERSION);
  printf("Copyright (C) 2005 -- 2007 Oohara Yuuma "
         "<oohara@libra.interq.or.jp>\n");
  printf("This program is free software;\n"
         "see the file LICENSE in the source code for details.\n");
  printf("two-minute-hate comes with NO WARRANTY, "
         "to the extent permitted by law.\n");
}
