/* hi emacs, this is a part of -*- C -*- source */
/* $Id: markov_lexical.l,v 1.20 2007/05/17 01:50:41 oohara Exp $ */
/* process this file with flex */

/* definitions */
%option prefix="markov_"
%option reentrant
%option noyywrap
%option nounput
%option noyyget_lineno
%option noyyget_in
%option noyyget_out
%option noyyget_leng
%option noyyset_lineno
%option noyyset_in
%option noyyset_out
%option noyyget_debug
%option noyyset_debug

%{
#include <stdio.h>

#include "markov.h"
#include "dbnj_util.h"

#include "markov_lexical.h"

#define YY_DECL int markov_lex(int n_errfile, FILE **p_errfile, \
  markov **chain, char **s, void *yyscanner)

/* to avoid gcc warnings */
int markov_get_column(yyscan_t yyscanner);
void markov_set_column(int column_no, yyscan_t yyscanner);
%}

/* rules */
/* this comment must be before %% */
%%
[a-zA-Z]([\-a-zA-Z\']?[a-zA-Z]+)*[.,;:]? {
  if (chain == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "markov_lex: chain is NULL\n");
    return 1;
  }
  if (s == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "markov_lex: s is NULL\n");
    return 1;
  }
  if (markov_get_text(yyscanner) == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "markov_lex: markov_get_text returned NULL\n");
    return 1;
  }

  if (*s != NULL)
  {
    if (markov_insert(n_errfile, p_errfile,
                      *chain,
                      *s, markov_get_text(yyscanner)) != 0)
    {
      dbnj_fprintf(n_errfile, p_errfile,
                   "markov_lex: markov_insert failed\n");
      return 1;
    }
  }
  if (*s != NULL)
    free(*s);
  *s = dbnj_concat_string(n_errfile, p_errfile,
                          NULL, 1, markov_get_text(yyscanner));
} 

.|\n {
  /* catch-all */
  /* ignore */;
}

<<EOF>> {
  if (*s != NULL)
    free(*s);
  return 0;
}

%%
/* user code */

void
markov_read_from_file(FILE *file, void *yyscanner)
{
  /* yyrestart */
  markov_restart(file, yyscanner);
}
