/* $Id: perlin.c,v 1.4 2005/12/07 15:10:47 oohara Exp $ */

#include <stdio.h>
/* rand, malloc */
#include <stdlib.h>

#include "perlin.h"

static int rect_average(int x_size, int y_size, int x, int y,
                        int c00, int cm0, int c0m, int cmm);

static int
rect_average(int x_size, int y_size, int x, int y,
             int c00, int cm0, int c0m, int cmm)
{
  if ((x_size <= 0) || (y_size <= 0))
    return c00;

  return (c00 * (x_size - x) * (y_size - y) + cm0 * x * (y_size - y)
          + c0m * (x_size - x) * y + cmm * x * y + x_size * y_size / 2)
    / (x_size * y_size);
}

/* each element of the returned array has a random value
 * in [0, loop * (loop + 1) / 2]
 * return NULL on error
 */
int *
perlin_noise(int (*suffix_func)(int, int, int, int),
             int x_size, int y_size, int loop)
{
  int x;
  int y;
  int step;
  int k;
  int x0;
  int y0;
  int xm;
  int ym;
  int c;
  int *noise = NULL;
  int *temp = NULL;

  if ((x_size <= 0) || (y_size <= 0))
    return NULL;
  if (loop < 0)
    return NULL;

  noise = (int *) malloc(sizeof(int) * x_size * y_size);
  if (noise == NULL)
    return NULL;

  temp = (int *) malloc(sizeof(int) * x_size * y_size);
  if (temp == NULL)
  {
    free(noise);
    noise = NULL;
    return NULL;
  }

  for (x = 0; x < x_size; x++)
    for (y = 0; y < y_size; y++)
      noise[suffix_func(x_size, y_size, x, y)] = 0;

  step = 1;
  for (k = 0; k < loop; k++)
  {
    for (x = 0; x < x_size; x += step)
    {
      for (y = 0; y < y_size; y += step)
      {
        c = rand() % 2;
        temp[suffix_func(x_size, y_size, x, y)] = c * (k + 1);
      }
    }

    for (x = 0; x < x_size; x++)
    {
      for (y = 0; y < y_size; y++)
      {
        x0 = (x / step) * step;
        xm = x0 + step;
        if (xm >= x_size)
          xm = 0;
        y0 = (y / step) * step;
        ym = y0 + step;
        if (ym >= y_size)
          ym = 0;
        noise[suffix_func(x_size, y_size, x, y)]
          += rect_average(step, step,
                          x % step,
                          y % step,
                          temp[suffix_func(x_size, y_size, x0, y0)],
                          temp[suffix_func(x_size, y_size, xm, y0)],
                          temp[suffix_func(x_size, y_size, x0, ym)],
                          temp[suffix_func(x_size, y_size, xm, ym)]);
      }
    }
    step *= 2;
    if ((step >= x_size) || (step >= y_size))
      break;
  }

  free(temp);
  temp = NULL;

  return noise;
}
